/*
 * The following source code is an example of how the
 * optical properties of milk are computed using the 
 * Lorenz-Mie theory presented in the SIGGRAPH 2007 paper
 *
 * Computing the Scattering Properties of Participating
 * Media Using Lorenz-Mie Theory
 *
 * By authors: Jeppe Revall Frisvad
 *             Niels Joergen Christensen
 *             Henrik Wann Jensen
 *
 * This paper holds full references to the literature 
 * mentioned in comments below.
 *
 * Code written by Jeppe Revall Frisvad, 2007.
 * Copyright (c) Jeppe Revall Frisvad 2007
 *
 * Permission is granted to anyone to use this code as 
 * software for any purpose, including commercial applications.
 * However, the software is provided 'as-is' without any warranty.
 */ 

#include <iostream>
#include <complex>

#include "LorenzMie.h"
#include "spectrum2rgb.h"

using namespace std;
using namespace LorenzMie;

namespace
{
  struct LogNormalParticleDistrib : public ParticleDistrib
  {
    double r_vs;  // volume-to-surface mean particle radius
    double c_s;   // coefficient of variation
  };

  // Data from Hale and Querry 1973
  complex<double> refrac_water[] = { complex<double>(1.341, 3.393e-10),
                     complex<double>(1.339, 2.110e-10),
                     complex<double>(1.338, 1.617e-10),
                     complex<double>(1.337, 3.302e-10),
                     complex<double>(1.336, 4.309e-10),
                     complex<double>(1.335, 8.117e-10),
                     complex<double>(1.334, 1.742e-9),
                     complex<double>(1.333, 2.473e-9),
                     complex<double>(1.333, 3.532e-9),
                     complex<double>(1.332, 1.062e-8),
                     complex<double>(1.332, 1.410e-8),
                     complex<double>(1.331, 1.759e-8),
                     complex<double>(1.331, 2.406e-8),
                     complex<double>(1.331, 3.476e-8),
                     complex<double>(1.330, 8.591e-8),
                     complex<double>(1.330, 1.474e-7),
                     complex<double>(1.330, 1.486e-7)  };

  // Absorbance from Du et al. 1998 (PhotochemCAD 1)
  double abs_riboflavin[] = { 0.07117, 0.05934, 0.07215, 0.08300,
                  0.06379, 0.01962, 0.00950, 0.00957,
                  0.01198, 0.01047, 0.01046, 0.01061,
                  0.00671, 0.01034, 0.00589, 0.00886, 0.01015 };

  // Data from Michalski et al. 2001
  complex<double> refrac_fat[] = { complex<double>(1.480, 4.0e-6),
                   complex<double>(1.476, 6.4e-6),
                   complex<double>(1.473, 8.6e-6),
                   complex<double>(1.470, 1.14e-5),
                   complex<double>(1.468, 1.10e-5),
                   complex<double>(1.466, 1.00e-5),
                   complex<double>(1.464, 4.7e-6),
                   complex<double>(1.463, 4.6e-6),
                   complex<double>(1.462, 4.7e-6),
                   complex<double>(1.461, 4.9e-6),
                   complex<double>(1.460, 5.0e-6),
                   complex<double>(1.459, 5.0e-6),
                   complex<double>(1.458, 5.1e-6),
                   complex<double>(1.457, 5.2e-6),
                   complex<double>(1.457, 5.2e-6),
                   complex<double>(1.456, 5.2e-6),
                   complex<double>(1.456, 5.2e-6)  };
}

int main()
{
#ifndef M_PI
  const double M_PI = 3.14159265358979323846;
#endif

  const int NO_SAMPLES = 17;
  const double M_SQRT_2PI = sqrt(2.0*M_PI);
  const double fat_density = 0.915;
  const double casein_density = 1.11;
  const double milk_density = 1.030;

  // User parameters:
  double pressure = 0.0;
  double fat_weight = 0.5;      // [0,88]
  double protein_weight = 3.4;  // [0,141]

  LogNormalParticleDistrib fat, casein;
  valarray<double> scatter(NO_SAMPLES);
  valarray<double> absorp(NO_SAMPLES);
  valarray<double> extinct(NO_SAMPLES);
  valarray<double> asymmetry(NO_SAMPLES);
  double volume;
  int counter;

  // Conversion from absorbance to absorption coefficients.
  // Scale based on molar absorption coefficient: 33000 M^-1 cm^-1 for 266.5 nm (Koziol 1966)
  double scale = 33.0e5/0.25654;
  for(int i = 0; i < NO_SAMPLES; ++i)
    abs_riboflavin[i] = scale*abs_riboflavin[i];

  // Weight fraction of riboflavin in milk (Fox and McSweeney 1998).
  // Natural content: 0.17 mg pr. 100 g.
  // Molar mass of riboflavin (C_17 H_20 N_4 O_6): 376.3682 g/mol
  const double w_riboflavin = log(10.0)*(0.17/376.3682)/(100.0/milk_density);

  // Correction of host medium refractive index
  counter = 0;
  for(double wavelength = 375.0e-9; wavelength < 780.0e-9; wavelength += 25.0e-9)
  {
    refrac_water[counter] += complex<double>(0.0, w_riboflavin*abs_riboflavin[counter]*wavelength/(4.0*M_PI));
    //cerr << abs_riboflavin[counter]
    //     << " " << refrac_water[counter].imag() << endl;
    ++counter;
  }

  // Functional expresions finding fat globule size distribution
  {
    const double log_50 = log10(50.0);
    const double v_sqr = fat_weight*fat_weight;
    const double r_P50 = 7.481403922e-4*v_sqr + 3.203991829e-2*fat_weight + 0.1162155667;
    const double r_P0 = fat_weight < 2.0 
                        ? -0.2528033794*v_sqr + 1.418606759*fat_weight 
                        : 1.456189142*pow(fat_weight, 0.36);
    const double beta = fat_weight > 0.0 ? log10(r_P0) : 0.0;
    const double alpha = beta/log_50 - log10(r_P50)/log_50;

    const double cv = 0.6;
    fat.r_vs = (pressure > 0.0 ? pow(10.0, -alpha*log10(pressure) + beta) : r_P0)/(cv*cv + 1.0);
    fat.c_s = fat.r_vs*cv;
    fat.r_vs *= 1.0e-6;
  }
  cerr << "Fat globule mean radius: " << fat.r_vs << endl;

  // More input data:
  fat.refrac_idx = complex<double>(1.462, 0.0);
  casein.refrac_idx =  complex<double>(1.503, 0.0);

  fat.r_min = 0.005e-6;
  fat.r_max = 10.0e-6;
  fat.dr = 0.05e-6;
  casein.r_vs = 43.0e-9;
  casein.c_s = 0.23;
  casein.r_min = 0.0e-9;
  casein.r_max = 150.0e-9;
  casein.dr = 1.0e-9;

  double fat_volume = (fat_weight/fat_density)/(100.0/milk_density);
  double casein_volume = 0.76*(protein_weight/casein_density)/(100.0/milk_density);
  cerr << "Fat volume fraction: " << fat_volume << endl;
  cerr << "Casein volume fraction: " << casein_volume << endl;

  fat.N.resize(static_cast<unsigned int>((fat.r_max - fat.r_min)/fat.dr) + 1);

  double alpha, beta, beta_sqr;
  if(!(fat_volume > 0.0))
  {
    alpha = 0.0;
    beta = 1.0;
  }
  else
  {
    beta_sqr = log(fat.c_s*fat.c_s/(fat.r_vs*fat.r_vs*1.0e12) + 1.0);
    alpha = log(fat.r_vs*1.0e6) - 0.5*beta_sqr;
    beta = sqrt(beta_sqr);
  }
  //cerr << "Fat alpha: " << alpha << " beta: " << beta << endl;
  volume = 0.0;
  counter = 0;
  for(double r = fat.r_min + fat.dr*0.5; r < fat.r_max; r += fat.dr)
  {
    double tmp = (log(r*1.0e6) - alpha)/beta;
    fat.N[counter] = 1.0/(r*1.0e6*beta*M_SQRT_2PI)*exp(-0.5*tmp*tmp);
    volume += fat.N[counter];
    ++counter;
  }
  fat.N *= fat_volume/volume;

  counter = 0;
  for(double r = fat.r_min + fat.dr*0.5; r < fat.r_max; r += fat.dr)
    fat.N[counter++] *= 3.0/(4.0*M_PI*r*r*r)*1.0/fat.dr;
  cerr << "Total number density of fat globules: " << fat.N.sum()*(fat.r_max - fat.r_min) << endl;

  casein.N.resize(static_cast<unsigned int>((casein.r_max - casein.r_min)/casein.dr) + 1);

  double x_vs = casein.r_vs*1.0e9/(150.0 - casein.r_vs*1.0e9);
  beta_sqr = log(casein.c_s*casein.c_s/(x_vs*x_vs) + 1.0);
  alpha = log(x_vs) - 0.5*beta_sqr;
  beta = sqrt(beta_sqr);
  //cerr << "Casein alpha: " << alpha << " beta: " << beta << endl;
  volume = 0.0;
  counter = 0;
  for(double r = casein.r_min + casein.dr*0.5; r < casein.r_max; r += casein.dr)
  {
    double x = r*1.0e9/(150.0 - r*1.0e9);
    double tmp = (log(x) - alpha)/beta;
    casein.N[counter] = 1.0/(x*beta*M_SQRT_2PI)*exp(-0.5*tmp*tmp);
    volume += casein.N[counter];
    ++counter;
  }
  casein.N *= casein_volume/volume;

  counter = 0;
  for(double r = casein.r_min + casein.dr*0.5; r < casein.r_max; r += casein.dr)
  {
    casein.N[counter] *= 3.0/(4.0*M_PI*r*r*r)*1.0/casein.dr;
    ++counter;
  }
  cerr << "Total number density of casein micelles: " << casein.N.sum()*(casein.r_max - casein.r_min) << endl << endl;

  for(double wavelength = 375.0e-9; wavelength < 780.0e-9; wavelength += 25.0e-9)
  {
    unsigned int wave_idx = static_cast<unsigned int>((wavelength - 375.0e-9)/24.9e-9);
    cerr << "Wavelength: " << wavelength << " (idx: " << wave_idx << ")" << endl;
    cerr << "Refractive Index of water: " << refrac_water[wave_idx] << endl;

    fat.ext = 0.0;
    fat.sca = 0.0;
    fat.abs = 0.0;
    fat.g = 0.0;
    counter = 0;
    for(double r = fat.r_min + fat.dr*0.5; r < fat.r_max; r += fat.dr)
    {
      double C_t, C_s, C_a, g, ior;
      particle_props(C_t, C_s, C_a, g, ior,
             r, wavelength, refrac_water[wave_idx], refrac_fat[wave_idx]);

      //cout << "Number density of fat globules at radius " << r << ": " << fat.N[counter] << endl;

      double sigma_s = C_s*fat.N[counter]*fat.dr;
      fat.ext += C_t*fat.N[counter]*fat.dr;
      fat.sca += sigma_s;
      fat.abs += C_a*fat.N[counter]*fat.dr;
      fat.g += g*sigma_s;
      ++counter;

      //cout << C_t << " " << C_s << " " << C_a << " " << g << endl;
      //cout << fat.ext << " " << fat.sca << " " << fat.abs << " " << fat.g << endl;
    }
    if(fat.sca != 0.0)
      fat.g /= fat.sca;
    cout << "Fat properties: " << fat.ext << " " << fat.sca << " " << fat.abs << " " << fat.g << endl;

    casein.ext = 0.0;
    casein.sca = 0.0;
    casein.abs = 0.0;
    casein.g = 0.0;
    counter = 0;
    for(double r = casein.r_min + casein.dr*0.5; r < casein.r_max; r += casein.dr)
    {
      double C_t, C_s, C_a, g, ior;
      particle_props(C_t, C_s, C_a, g, ior,
             r, wavelength, refrac_water[wave_idx], casein.refrac_idx);

      //cout << "Number density of casein micelles at radius " << r << ": " << casein.N[counter] << endl;

      double sigma_s = C_s*casein.N[counter]*casein.dr;
      casein.ext += C_t*casein.N[counter]*casein.dr;
      casein.sca += sigma_s;
      casein.abs += C_a*casein.N[counter]*casein.dr;
      casein.g += g*sigma_s;
      ++counter;

      //cout << C_t << " " << C_s << " " << C_a << " " << g << endl;
      //cout << casein.ext << " " << casein.sca << " " << casein.abs << " " << casein.g << endl;
    }
    if(casein.sca != 0.0)
      casein.g /= casein.sca;
    cout << "Casein properties: " << casein.ext << " " << casein.sca
     << " " << casein.abs << " " << casein.g << endl;

    double water_absorp = 4.0*M_PI*refrac_water[wave_idx].imag()/wavelength;
    cerr << "water extinction: " << water_absorp << endl;
    if(casein.sca - casein.ext > water_absorp)
    {
      cerr << "Lorenz-Mie coefficients did not converge or integration is too coarse." << endl;
      exit(0);
    }

    extinct[wave_idx] = fat.ext + casein.ext + water_absorp;
    scatter[wave_idx] = fat.sca + casein.sca;
    absorp[wave_idx] = extinct[wave_idx] - scatter[wave_idx];
    if(scatter[wave_idx] != 0.0)
      asymmetry[wave_idx] = (fat.sca*fat.g + casein.sca*casein.g)/scatter[wave_idx];

    cout << "Extinction coefficient (" << wavelength*1e9 << "nm): " << extinct[wave_idx] << endl
     << "Scattering coefficient (" << wavelength*1e9 << "nm): " << scatter[wave_idx] << endl
     << "Absorption coefficient (" << wavelength*1e9 << "nm): " << absorp[wave_idx] << endl
     //<< "Fat asymmetry parameter (" << wavelength*1e9 << "nm):    " << fat.g << endl
     //<< "Casein asymmetry parameter (" << wavelength*1e9 << "nm):    " << casein.g << endl
     << "Ensemble asymmetry parameter (" << wavelength*1e9 << "nm):    " << asymmetry[wave_idx] << endl;

    if(wavelength < 774.0e-9)
    {
      cout << endl << "<press enter for next set of values>" << endl;
      cin.get();
    }
  }

  valarray<double> RGB[3];
  for(int i = 0; i < 3; ++i)
    RGB[i].resize(NO_SAMPLES - 1);

  for(int i = 0; i < NO_SAMPLES - 1; ++i)
    for(int j = 0; j < 3; ++j)
      RGB[j][i] = spectrum_rgb[i*5 + 2][j + 1];

  for(int i = 0; i < 3; ++i)
    RGB[i] /= RGB[i].sum();

  double rgb_extinct[3];
  double rgb_scatter[3];
  double rgb_absorp[3];
  double rgb_asymmetry[3];
  complex<double> rgb_water_ior[3];
  double rgb_water_absorp[3];
  for(int j = 0; j < 3; ++j)
  {
    rgb_extinct[j] = 0.0;
    rgb_scatter[j] = 0.0;
    rgb_absorp[j] = 0.0;
    rgb_asymmetry[j] = 0.0;
    rgb_water_ior[j] = complex<double>(0.0, 0.0);
    rgb_water_absorp[j] = 0.0;
  }

  for(int i = 0; i < NO_SAMPLES - 1; ++i)
    for(int j = 0; j < 3; ++j)
    {
      rgb_extinct[j] += RGB[j][i]*extinct[i + 1];
      rgb_scatter[j] += RGB[j][i]*scatter[i + 1];
      rgb_absorp[j]  += RGB[j][i]*absorp[i + 1];
      rgb_asymmetry[j] += RGB[j][i]*asymmetry[i + 1];
      rgb_water_ior[j] += RGB[j][i]*refrac_water[i + 1];
      double wavelength = 400.0e-9 + i*25.0e-9;
      rgb_water_absorp[j] += RGB[j][i]*4.0*M_PI*refrac_water[i + 1].imag()/wavelength;
    }

  cout << endl << "RGB values corresponding to the spectral optical properties:" << endl;
  cout << "Extinction coefficient: (" << rgb_extinct[0]
       << ", " << rgb_extinct[1] << ", " << rgb_extinct[2] << ")" << endl
       << "Scattering coefficient: (" << rgb_scatter[0]
       << ", " << rgb_scatter[1] << ", " << rgb_scatter[2] << ")" << endl
       << "Absorption coefficient: (" << rgb_absorp[0] << ", "
       << rgb_absorp[1] << ", " << rgb_absorp[2] << ")" << endl
       << "Asymmetry parameter:    (" << rgb_asymmetry[0] << ", "
       << rgb_asymmetry[1] << ", " << rgb_asymmetry[2] << ")" << endl;

  cout << endl << "RGB values for water with vitamin B2 (Riboflavin):" << endl
       << "Absorption coefficient: (" << rgb_water_absorp[0]
       << ", " << rgb_water_absorp[1] << ", " << rgb_water_absorp[2] << ")" << endl
       << "Index of refraction:    [ " << rgb_water_ior[0] << "," << endl 
       << "                          " << rgb_water_ior[1] << "," << endl
       << "                          " << rgb_water_ior[2] << " ]";

  return 0;
}
